@@ -195,7 +195,7 @@ static void save_cookies(ghttp_request *request)
 int propfind(ghttp_request *request, char *myurl, int whichrequest)
 {
 	int status;
-	
+
 	/* Settings of our connection	 */
 	ghttp_clean(request);
 	ghttp_set_uri(request, myurl);
@@ -247,7 +247,9 @@ int propfind(ghttp_request *request, char *myurl, int whichrequest)
 	  set_authinfo(request, UserName, PassWord, WWWAuthenticate, 0);
 
 	}
-	if (whichrequest == MSGLIST)
+
+	/* TODO: Check if we have return number 51 which is password requried before nulling out auth */
+	if (whichrequest == MSGLIST && !auth_required)
 	  set_authinfo(request, UserName, PassWord, NULL, 1);
 
 	/* now prepare and send our http request that we've built */
@@ -584,7 +586,72 @@ int update_readmark(MSG_STRUCT *msg_props, int new_readmark) {
 
   return (reply == OKSTATUS || reply == MULTISTATUS) ? E_OK : E_UPDATE_FAILED;
 }
-		    
+
+void setnewsmtpurl(void) {
+    if (smtpurl->href) {
+	free(smtpurl->href);
+	smtpurl->href = NULL;
+    }
+    smtpurl->href = get_moved_url(request);
+}
+
+int handlesmtpreplies(int reply) {
+    const char *wwwauth;
+
+    if (reply != AUTHORIZATION)
+	unauth_cnt=0; /* reset the unauthorised counter */
+    if (reply == AUTHORIZATION) {
+	if ((wwwauth = ghttp_get_header(request, http_hdr_WWW_Authenticate)) != NULL) {
+	    if (WWWAuthenticate)
+		free(WWWAuthenticate);
+	    WWWAuthenticate = strdup(wwwauth);
+	    auth_required = 1;
+#ifdef DEBUG
+	    fprintf( stderr, "reply=AUTHORIZATION; %s\n", wwwauth );
+#endif
+	}
+	return 0;
+    }
+    if (reply == MOVED)
+	return 0;
+    return reply;
+}
+
+/* handling message posting, mainly for hotsmtpd */
+int post(char *body, unsigned long body_len) {
+    int status, ret_len;
+
+    ghttp_clean(request);
+    ghttp_set_uri(request, smtpurl->href);
+    if (ghttp_set_type(request, ghttp_type_post) == -1) 
+	return E_GHTTP_ERR;
+    if (ghttp_set_body(request, body, body_len) == -1)
+	return E_GHTTP_ERR;
+    /* Do not go wildly adding headers.  httpmail servers are very picky *
+     * about what headers they see.  If this function stops working properly *
+     * someone went screwing around with headers here or in prepare_and_send. *
+     * If this isn't it, then httpmail has changed its interface! */  
+    ghttp_set_header(request, "Content-Type", "message/rfc821");
+    set_authinfo(request, UserName, PassWord, NULL, 1);
+
+    prepare_and_send(request,smtpurl->href);
+
+    status = ghttp_status_code(request);
+    save_cookies(request);
+
+    if (status == AUTHORIZATION) {
+	int i;
+	if (unauth_cnt++ > 3) /* make it 4 times just to be sure :-) */
+	    return PASSWORDFAILED;
+	for (i = 0; i < badpwdcnt; i++) {
+	    if (!strcmp(badpwdtxt[i], ghttp_reason_phrase(request))) 
+		return PASSWORDFAILED;
+	}
+    } else
+	auth_required = 0;
+
+    return status;
+}
 
 static int proppatch(char *href, char *xml_request) {
   int status, len = strlen(xml_request);
@@ -661,9 +728,18 @@ void httpmail_destroy(void)
 	while(cookiecount>0) {
 		cookiecount--;
 		if( cookies[cookiecount] ){
-			free(cookies[cookiecount]->key);
-			free(cookies[cookiecount]->value);
-			free(cookies[cookiecount]->domain);
+			if ( cookies[cookiecount]->key ) { 
+				free(cookies[cookiecount]->key);
+				cookies[cookiecount]->key = NULL;
+			}
+			if ( cookies[cookiecount]->value ) {
+				free(cookies[cookiecount]->value);
+				cookies[cookiecount]->value = NULL;
+			}
+			if ( cookies[cookiecount]->domain ) {
+				free(cookies[cookiecount]->domain);
+				cookies[cookiecount]->domain = NULL;
+			}
 			free(cookies[cookiecount]);
 			cookies[cookiecount]= NULL;
 		}
@@ -961,9 +1037,12 @@ static int allowed_user(char *user) {
 
 char *grow_and_copy(char *dest, int curpos, char *src, int i)
 {
+  if (dest==NULL) curpos=0;
+
   if(!(dest = realloc(dest, curpos+i+1)))
     return NULL;
 
+  dest[curpos]='\0'; /* dest must be null terminated */
   strlcat(dest, src, curpos+i+1);
 
   return dest;
