@@ -42,6 +42,7 @@
 
 static int state; /* current smtp state */
 static int timeout, timed_out = 0;
+static char auth_done = 0; /* flag if authorzation is done */
 
 /* Global variables */
 sasl_conn_t *sasl_conn;
@@ -75,10 +76,10 @@ void get_input(void)
     if (errno != EINTR) {
       timed_out = errno;
       alarm(0);
-      strlcpy(input, "timeout\n", N_BUFLEN);
-    } else { 
+      strlcpy(input, "TIMEOUT\n", N_BUFLEN);
+    } else {
       if (timed_out) {
-	strlcpy(input, "timeout\n", N_BUFLEN);
+	strlcpy(input, "TIMEOUT\n", N_BUFLEN);
       }
     }
   }
@@ -132,7 +133,7 @@ int do_login(sasl_conn_t *conn, void *context, const char *uname, const char *pw
     }
     break;
   default:
-    PFSOUT("451 httpmail login BAD %s[%d] Internal error", __FILE__, __LINE__);
+    PFSOUT("421 httpmail login BAD %s[%d] Internal error", __FILE__, __LINE__);
     PCRLF;
     return SASL_NOUSER;
   }
@@ -163,7 +164,7 @@ int do_login(sasl_conn_t *conn, void *context, const char *uname, const char *pw
   case ERROR:
     if (log_level > 0)
       LOG("AUTH failed, unexpected response from server");
-    PSOUT("451 Remote server returned unexpected status");
+    PSOUT("421 Remote server returned unexpected status");
     PCRLF;
     return SASL_NOUSER;
   case 0:
@@ -194,6 +195,12 @@ static void hello_cmd( cmdtable *c, int argc, char *argv[])
   int result;
   const char *result_string;
   char supported[100]={'\0'};
+  char have_sasl_mech;
+
+  delete_list(&HEADERS);
+  delete_list(&MAILDATA);
+
+  if(argc < 2) goto no_domain;
 
   if(strlen(argv[1]) > MAX_DOMAIN_LEN)
     {
@@ -202,7 +209,6 @@ static void hello_cmd( cmdtable *c, int argc, char *argv[])
       return;
     }
 
-  if(argc < 1) goto no_domain;
   if(strlen(argv[1]) < 3) 
     {
     no_domain:
@@ -212,26 +218,64 @@ static void hello_cmd( cmdtable *c, int argc, char *argv[])
       return;
     }
 
-  /* Make a new context for this connection */
-  result=sasl_server_new("hotsmtpd", NULL, NULL, NULL, NULL, NULL, 0, &sasl_conn);
+  if (!auth_done)
+  {
+    /* Only needed first time when client is not yet authenticated */
 
-  result=sasl_listmech(sasl_conn, NULL, NULL, " ", NULL, &result_string, NULL, NULL);
-	
-  if(strstr(result_string, "LOGIN")) strcpy(supported, "LOGIN ");
-  if(strstr(result_string, "PLAIN")) strcat(supported, "PLAIN");
+    have_sasl_mech=0;
+    /* Flag to see if we have any usable sasl authentication mechanism
+       available in the cyrus-sasl implementation on this machine */
+
+    /* Make a new context for this connection */
+    result=sasl_server_new("hotsmtpd", NULL, NULL, NULL, NULL, NULL, 0, &sasl_conn);
+
+    result=sasl_listmech(sasl_conn, NULL, NULL, " ", NULL, &result_string, NULL, NULL);
+
+    if(strstr(result_string, "LOGIN"))
+    {
+      strcpy(supported, "LOGIN ");
+      have_sasl_mech=1;
+    }
+    if(strstr(result_string, "PLAIN"))
+    {
+      strcat(supported, "PLAIN");
+      have_sasl_mech=1;
+    }
+
+    if (!have_sasl_mech)
+    {
+      /* Neither PLAIN nor LOGIN authentication mechanism are supported
+	 in the cyrus-sasl library on this computer. We are not able to
+	 service any requests. */
+      PSOUT("451-Failed finding an authentication mechanism in cyrus-sasl library");
+      PCRLF;
+      PSOUT("451 Service not available, closing transmission channel");
+      PCRLF;
+      state = TERMINATE;
+      return;
+    }
+  }
 
   PFSOUT("250-%s Pleased to meet you", argv[1]);
   PCRLF;
-  if(strcasecmp(argv[0], "EHLO")==0) {
-    /* EHLO is ESMTP and needs to show supported authentication and extensions */ 
-    PFSOUT("250-AUTH %s", supported);
-    PCRLF;
-    PFSOUT("250-AUTH=%s", supported);
-    PCRLF;
+  if(strcasecmp(argv[0], "EHLO")==0)
+  {
+    /* EHLO is ESMTP and needs to show supported authentication and extensions */
+    if (!auth_done)
+    {
+      PFSOUT("250-AUTH %s", supported);
+      PCRLF;
+      PFSOUT("250-AUTH=%s", supported);
+      PCRLF;
+    }
     PSOUT("250 8BITMIME");
     PCRLF;
   }
-  state = HELLO;
+  if (auth_done) {
+    state = AUTHENTICATED;
+  } else {
+    state = HELLO;
+  }
 }
 
 /* Many want better security... sorry.  We need plain text passwords and user
@@ -241,12 +285,37 @@ static void auth_cmd( cmdtable *c, int argc, char *argv[])
 {
   int result;
   char var64[MAX_COMMAND_LEN]={'\0'};
+  char auth_method[MAX_COMMAND_LEN] = {'\0'};
   unsigned var64len;
   const char *out;
   unsigned outlen;
 
-  sasl_decode64(argv[2], strlen(argv[2]), var64, MAX_COMMAND_LEN, &var64len);
-  result=sasl_server_start(sasl_conn, argv[1], var64, var64len, &out, &outlen);
+  if (auth_done)
+  {
+      PSOUT("503 No more authentications can be done in this session");
+      PCRLF;
+      return;
+  }
+  if (argc<2)
+  {
+      PSOUT("501 Authentication failed!");
+      PCRLF;
+      return;
+  }
+  else if (argc == 2) /* this is needed to get kmail to work.. */
+  {
+      strncpy(auth_method, argv[1], MAX_COMMAND_LEN);
+      PSOUT("334");
+      PCRLF;
+      get_input();
+      sasl_decode64(input, strlen(input), var64, MAX_COMMAND_LEN, &var64len);
+      result=sasl_server_start(sasl_conn, auth_method, var64, var64len, &out, &outlen);
+  }
+  else /* argc >= 3 */
+  { 
+      sasl_decode64(argv[2], strlen(argv[2]), var64, MAX_COMMAND_LEN, &var64len);
+      result=sasl_server_start(sasl_conn, argv[1], var64, var64len, &out, &outlen);
+  }
 
   while(result == SASL_CONTINUE) /* Fall through for anything but a continue */
     {
@@ -261,11 +330,12 @@ static void auth_cmd( cmdtable *c, int argc, char *argv[])
       sasl_decode64(input, strlen(input), var64, MAX_COMMAND_LEN, &var64len);
       result=sasl_server_step(sasl_conn, var64, var64len, &out, &outlen);
     }
-  sasl_dispose(&sasl_conn);
   if (result == SASL_OK)
     {
       PSOUT("235 Authentication successful!");
       PCRLF;
+      sasl_dispose(&sasl_conn);
+      auth_done=1;
       state = AUTHENTICATED;
       return;
     }
@@ -287,7 +357,7 @@ static void logout_cmd( cmdtable *c, int argc, char *argv[])
 {
   PSOUT("221 Service closing transmission channel");
   PCRLF;
-  if (state == AUTHENTICATED) {
+  if (auth_done) {
     delete_list(&HEADERS);
     delete_list(&MAILDATA);
     state = LOGOUT;
@@ -299,9 +369,14 @@ static void logout_cmd( cmdtable *c, int argc, char *argv[])
 static void rset_cmd( cmdtable *c, int argc, char *argv[])
 {
   delete_list(&HEADERS);
+  delete_list(&MAILDATA);
   PSOUT("250 OK");
   PCRLF;
-  state = HELLO;
+  if (auth_done) {
+    state = AUTHENTICATED;
+  } else {
+    state = HELLO;
+  }
 }
 
 static void data_cmd( cmdtable *c, int argc, char *argv[])
@@ -316,11 +391,13 @@ static void data_cmd( cmdtable *c, int argc, char *argv[])
 
 void mail_cmd( cmdtable *c, int argc, char *argv[])
 {
-  if(strncasecmp(argv[1], "FROM:", 5)!=0) {
+
+  if(argc < 2 || strncasecmp(argv[1], "FROM:", 5)!=0) {
     PSOUT("501 Syntax: MAIL FROM: <address>");
     PCRLF;
     return;
   }
+
   if(strlen(input)>MAX_COMMAND_LEN) {
     PSOUT("501 MAIL line too long");
     PCRLF;
@@ -335,7 +412,7 @@ void mail_cmd( cmdtable *c, int argc, char *argv[])
 
 void rcpt_cmd( cmdtable *c, int argc, char *argv[])
 {
-  if(strncasecmp(argv[1], "TO:", 3)!=0) {
+  if(argc < 2 || strncasecmp(argv[1], "TO:", 3)!=0) {
     PSOUT("501 Syntax: RCPT TO: <address>");
     PCRLF;
     return;
@@ -363,15 +440,16 @@ void exit_cmd( cmdtable *c, int argc, char *argv[])
  * Global jumptable for commands
  */
 const cmdtable commands[] = {
-  { "AUTH", (HELLO), &auth_cmd },
-  { "HELO", (NEW), &hello_cmd },
-  { "EHLO", (NEW), &hello_cmd },
+  { "AUTH", (HELLO | AUTHENTICATED), &auth_cmd },
+  { "HELO", (HELLO | AUTHENTICATED | RCPTTO | MAILFROM | NEW), &hello_cmd },
+  { "EHLO", (HELLO | AUTHENTICATED | RCPTTO | MAILFROM | NEW), &hello_cmd },
   { "MAIL", (AUTHENTICATED), &mail_cmd },
   { "RCPT", (MAILFROM | RCPTTO), &rcpt_cmd },
   { "DATA", (RCPTTO), &data_cmd },
-  { "RSET", (AUTHENTICATED | RCPTTO | MAILFROM), &rset_cmd },
-  { "NOOP", (AUTHENTICATED | RCPTTO | MAILFROM | NEW), &noop_cmd },
+  { "RSET", (HELLO | AUTHENTICATED | RCPTTO | MAILFROM | NEW), &rset_cmd },
+  { "NOOP", (HELLO | AUTHENTICATED | RCPTTO | MAILFROM | NEW), &noop_cmd },
   { "QUIT", (HELLO | AUTHENTICATED | RCPTTO | MAILFROM | NEW), &logout_cmd },
+  { "TIMEOUT", (HELLO | AUTHENTICATED | RCPTTO | MAILFROM | NEW), &exit_cmd },
   { "", 0, NULL }
 };
 
@@ -457,7 +535,7 @@ void main_loop(void)
 	if (c->states & state) {
 	  (c->handler)(c, argc, argv); /* run the command handler */
 	} else {
-	  if(!(state & AUTHENTICATED)) PFSOUT("503 Authenticate first! - %s command is not available right now.", argv[0]);
+	  if(!auth_done) PFSOUT("503 Authenticate first! - %s command is not available right now.", argv[0]);
 	  else PFSOUT("503 Bad sequence of commands - %s command is not available right now.", argv[0]);
 	  PCRLF;
 	}
@@ -518,7 +596,7 @@ int main(int argc, char **argv)
  
   if(init_sasl() != 0)
     {
-      PSOUT("451 Unable to initialize sasl library!");
+      PSOUT("451 Internal Error: Unable to initialize sasl library!");
       PCRLF;
       exit(-1);
     }
