@@ -670,13 +670,17 @@ static void cleanup_folder_props(FOLDER_STRUCT *folder_props)
 int ParseXMLMsgList(ghttp_request *request, FOLDER_STRUCT *folder_props)
 {
   int i, n;
+  char *src;
 
   if (folder_props->msg_props || folder_props->unread_props) {
     /* urgh, let's clean it this up */
     cleanup_folder_props(folder_props);
   }
-  
-  update_visiblecount(ghttp_get_body(request), folder_props);
+
+  src = ghttp_get_body(request);
+  remove_to_from_subject (src);
+
+  update_visiblecount(src, folder_props);
   
   if ((n = folder_props->visiblecount) > 0) {
 	/* Creates array of MSG_STRUCTs */
@@ -693,7 +697,7 @@ int ParseXMLMsgList(ghttp_request *request, FOLDER_STRUCT *folder_props)
 	
 	/* Fill our MSG_STRUCTS */
 	/* RJVB 20020927: necessary to update visiblecount and msg_props if the nr. of messages changed! */
-	msglengths(ghttp_get_body(request), folder_props);
+	msglengths(src, folder_props);
 	
 	/* Build the list of unread messages */
 	folder_props->unreadcount = buildunreadlist(folder_props->msg_props, n, folder_props->unread_props);
@@ -701,6 +705,55 @@ int ParseXMLMsgList(ghttp_request *request, FOLDER_STRUCT *folder_props)
   return 1;
 }
 
+/*#define DEBUG_HEAD 1*/
+
+/* remove to, from and subject contents because they may contain UTF-16 characters and those cause parsing error */
+remove_to_from_subject (char *in)
+{
+  char *p, *q, *r;
+  int i;
+  static const struct {
+    char *start;
+    char *end;
+  } eliminate[] = {
+    {"<m:subject>", "</m:subject>"},
+    {"<m:from>", "</m:from>"},
+    {"<m:to>", "</m:to>"},
+    {NULL, NULL}
+  };
+
+#ifdef DEBUG_HEAD
+  FILE *fp;
+
+  fp = fopen ("/tmp/111.txt", "a");
+  fprintf (fp, "1: %s\n", in);
+#endif
+
+  for (i=0; eliminate[i].start != NULL; i++)
+    {
+      for (r =in;; r = p+strlen(eliminate[i].end))
+        {
+          p = strstr (r, eliminate[i].start);
+          q = strstr (r, eliminate[i].end);
+          if (p != NULL && q != NULL)
+            {
+              p += strlen (eliminate[i].start);
+              while (p < q)
+                *p++ = ' ';
+            }
+          else
+            {
+              break;
+            }
+        }
+    }
+
+#ifdef DEBUG_HEAD
+  fprintf (fp, "2: %s\n", in);
+  fclose (fp);
+#endif
+}
+
 
 /* Use this function to attempt to find the folder foldername and fill in the
  * folder_props->href, folder_props->visiblecount and
